@@ -3,17 +3,26 @@
 
 #include "alloc.h"
 #include "bit.h"
+#include "config.h"
 #include "diag.h"
 #include "sanity.h"
 #include <errno.h>
 #include <stdlib.h>
+#include <stdint.h>
 #include <string.h>
 
+/** The largest possible allocation size. */
+#if PTRDIFF_MAX < SIZE_MAX / 2
+#  define ALLOC_MAX ((size_t)PTRDIFF_MAX)
+#else
+#  define ALLOC_MAX (SIZE_MAX / 2)
+#endif
+
 /** Portable aligned_alloc()/posix_memalign(). */
 static void *xmemalign(size_t align, size_t size) {
 	bfs_assert(has_single_bit(align));
 	bfs_assert(align >= sizeof(void *));
-	bfs_assert((size & (align - 1)) == 0);
+	bfs_assert(is_aligned(align, size));
 
 #if __APPLE__
 	void *ptr = NULL;
@@ -26,9 +35,9 @@ static void *xmemalign(size_t align, size_t size) {
 
 void *alloc(size_t align, size_t size) {
 	bfs_assert(has_single_bit(align));
-	bfs_assert((size & (align - 1)) == 0);
+	bfs_assert(is_aligned(align, size));
 
-	if (size >> (SIZE_WIDTH - 1)) {
+	if (size > ALLOC_MAX) {
 		errno = EOVERFLOW;
 		return NULL;
 	}
@@ -42,9 +51,9 @@ void *alloc(size_t align, size_t size) {
 
 void *zalloc(size_t align, size_t size) {
 	bfs_assert(has_single_bit(align));
-	bfs_assert((size & (align - 1)) == 0);
+	bfs_assert(is_aligned(align, size));
 
-	if (size >> (SIZE_WIDTH - 1)) {
+	if (size > ALLOC_MAX) {
 		errno = EOVERFLOW;
 		return NULL;
 	}
@@ -60,6 +69,66 @@ void *zalloc(size_t align, size_t size) {
 	return ret;
 }
 
+void *xrealloc(void *ptr, size_t align, size_t old_size, size_t new_size) {
+	bfs_assert(has_single_bit(align));
+	bfs_assert(is_aligned(align, old_size));
+	bfs_assert(is_aligned(align, new_size));
+
+	if (new_size == 0) {
+		free(ptr);
+		return NULL;
+	} else if (new_size > ALLOC_MAX) {
+		errno = EOVERFLOW;
+		return NULL;
+	}
+
+	if (align <= alignof(max_align_t)) {
+		return realloc(ptr, new_size);
+	}
+
+	// There is no aligned_realloc(), so reallocate and copy manually
+	void *ret = xmemalign(align, new_size);
+	if (!ret) {
+		return NULL;
+	}
+
+	size_t min_size = old_size < new_size ? old_size : new_size;
+	if (min_size) {
+		memcpy(ret, ptr, min_size);
+	}
+
+	free(ptr);
+	return ret;
+}
+
+void *reserve(void *ptr, size_t align, size_t size, size_t count) {
+	// No need to overflow-check the current size
+	size_t old_size = size * count;
+
+	// Capacity is doubled every power of two, from 0→1, 1→2, 2→4, etc.
+	// If we stayed within the same size class, re-use ptr.
+	if (count & (count - 1)) {
+		// Tell sanitizers about the new array element
+		sanitize_alloc((char *)ptr + old_size, size);
+		errno = 0;
+		return ptr;
+	}
+
+	// No need to overflow-check; xrealloc() will fail before we overflow
+	size_t new_size = count ? 2 * old_size : size;
+	void *ret = xrealloc(ptr, align, old_size, new_size);
+	if (!ret) {
+		// errno is used to communicate success/failure to the RESERVE() macro
+		bfs_assert(errno != 0);
+		return ptr;
+	}
+
+	// Pretend we only allocated one more element
+	sanitize_free((char *)ret + old_size + size, new_size - old_size - size);
+	errno = 0;
+	return ret;
+}
+
 /**
  * An arena allocator chunk.
  */
@@ -89,7 +158,7 @@ static void chunk_set_next(const struct arena *arena, union chunk *chunk, union
 
 void arena_init(struct arena *arena, size_t align, size_t size) {
 	bfs_assert(has_single_bit(align));
-	bfs_assert((size & (align - 1)) == 0);
+	bfs_assert(is_aligned(align, size));
 
 	if (align < alignof(union chunk)) {
 		align = alignof(union chunk);
@@ -97,7 +166,7 @@ void arena_init(struct arena *arena, size_t align, size_t size) {
 	if (size < sizeof(union chunk)) {
 		size = sizeof(union chunk);
 	}
-	bfs_assert((size & (align - 1)) == 0);
+	bfs_assert(is_aligned(align, size));
 
 	arena->chunks = NULL;
 	arena->nslabs = 0;
@@ -107,13 +176,8 @@ void arena_init(struct arena *arena, size_t align, size_t size) {
 }
 
 /** Allocate a new slab. */
+attr(cold)
 static int slab_alloc(struct arena *arena) {
-	void **slabs = realloc(arena->slabs, sizeof_array(void *, arena->nslabs + 1));
-	if (!slabs) {
-		return -1;
-	}
-	arena->slabs = slabs;
-
 	// Make the initial allocation size ~4K
 	size_t size = 4096;
 	if (size < arena->size) {
@@ -130,6 +194,13 @@ static int slab_alloc(struct arena *arena) {
 		return -1;
 	}
 
+	// Grow the slab array
+	void **pslab = RESERVE(void *, &arena->slabs, &arena->nslabs);
+	if (!pslab) {
+		free(slab);
+		return -1;
+	}
+
 	// Fix the last chunk->next offset
 	void *last = (char *)slab + size - arena->size;
 	chunk_set_next(arena, last, arena->chunks);
@@ -137,7 +208,7 @@ static int slab_alloc(struct arena *arena) {
 	// We can rely on zero-initialized slabs, but others shouldn't
 	sanitize_uninit(slab, size);
 
-	arena->chunks = arena->slabs[arena->nslabs++] = slab;
+	arena->chunks = *pslab = slab;
 	return 0;
 }
 
@@ -208,21 +279,16 @@ static size_t varena_exact_size(const struct varena *varena, size_t count) {
 static struct arena *varena_get(struct varena *varena, size_t count) {
 	size_t i = varena_size_class(varena, count);
 
-	if (i >= varena->narenas) {
-		size_t narenas = i + 1;
-		struct arena *arenas = realloc(varena->arenas, sizeof_array(struct arena, narenas));
-		if (!arenas) {
+	while (i >= varena->narenas) {
+		size_t j = varena->narenas;
+		struct arena *arena = RESERVE(struct arena, &varena->arenas, &varena->narenas);
+		if (!arena) {
 			return NULL;
 		}
 
-		for (size_t j = varena->narenas; j < narenas; ++j) {
-			size_t shift = j + varena->shift;
-			size_t size = varena_exact_size(varena, (size_t)1 << shift);
-			arena_init(&arenas[j], varena->align, size);
-		}
-
-		varena->narenas = narenas;
-		varena->arenas = arenas;
+		size_t shift = j + varena->shift;
+		size_t size = varena_exact_size(varena, (size_t)1 << shift);
+		arena_init(arena, varena->align, size);
 	}
 
 	return &varena->arenas[i];
