@@ -142,12 +142,12 @@ static int init_esc(struct colors *colors, const char *name, const char *value,
 	*field = esc;
 
 	struct trie_leaf *leaf = trie_insert_str(&colors->names, name);
-	if (leaf) {
-		leaf->value = field;
-		return 0;
-	} else {
+	if (!leaf) {
 		return -1;
 	}
+
+	leaf->value = field;
+	return 0;
 }
 
 /** Check if an escape sequence is equal to a string. */
@@ -623,58 +623,58 @@ struct colors *parse_colors(void) {
 	trie_init(&colors->ext_trie);
 	trie_init(&colors->iext_trie);
 
-	int ret = 0;
+	bool fail = false;
 
 	// From man console_codes
 
-	ret |= init_esc(colors, "rs", "0",      &colors->reset);
-	ret |= init_esc(colors, "lc", "\033[",  &colors->leftcode);
-	ret |= init_esc(colors, "rc", "m",      &colors->rightcode);
-	ret |= init_esc(colors, "ec", NULL,     &colors->endcode);
-	ret |= init_esc(colors, "cl", "\033[K", &colors->clear_to_eol);
+	fail = fail || init_esc(colors, "rs", "0",      &colors->reset);
+	fail = fail || init_esc(colors, "lc", "\033[",  &colors->leftcode);
+	fail = fail || init_esc(colors, "rc", "m",      &colors->rightcode);
+	fail = fail || init_esc(colors, "ec", NULL,     &colors->endcode);
+	fail = fail || init_esc(colors, "cl", "\033[K", &colors->clear_to_eol);
 
-	ret |= init_esc(colors, "bld", "01;39", &colors->bold);
-	ret |= init_esc(colors, "gry", "01;30", &colors->gray);
-	ret |= init_esc(colors, "red", "01;31", &colors->red);
-	ret |= init_esc(colors, "grn", "01;32", &colors->green);
-	ret |= init_esc(colors, "ylw", "01;33", &colors->yellow);
-	ret |= init_esc(colors, "blu", "01;34", &colors->blue);
-	ret |= init_esc(colors, "mag", "01;35", &colors->magenta);
-	ret |= init_esc(colors, "cyn", "01;36", &colors->cyan);
-	ret |= init_esc(colors, "wht", "01;37", &colors->white);
+	fail = fail || init_esc(colors, "bld", "01;39", &colors->bold);
+	fail = fail || init_esc(colors, "gry", "01;30", &colors->gray);
+	fail = fail || init_esc(colors, "red", "01;31", &colors->red);
+	fail = fail || init_esc(colors, "grn", "01;32", &colors->green);
+	fail = fail || init_esc(colors, "ylw", "01;33", &colors->yellow);
+	fail = fail || init_esc(colors, "blu", "01;34", &colors->blue);
+	fail = fail || init_esc(colors, "mag", "01;35", &colors->magenta);
+	fail = fail || init_esc(colors, "cyn", "01;36", &colors->cyan);
+	fail = fail || init_esc(colors, "wht", "01;37", &colors->white);
 
-	ret |= init_esc(colors, "wrn", "01;33", &colors->warning);
-	ret |= init_esc(colors, "err", "01;31", &colors->error);
+	fail = fail || init_esc(colors, "wrn", "01;33", &colors->warning);
+	fail = fail || init_esc(colors, "err", "01;31", &colors->error);
 
 	// Defaults from man dir_colors
 	// "" means fall back to ->normal
 
-	ret |= init_esc(colors, "no", NULL,    &colors->normal);
+	fail = fail || init_esc(colors, "no", NULL,    &colors->normal);
 
-	ret |= init_esc(colors, "fi", "",      &colors->file);
-	ret |= init_esc(colors, "mh", NULL,    &colors->multi_hard);
-	ret |= init_esc(colors, "ex", "01;32", &colors->executable);
-	ret |= init_esc(colors, "ca", NULL,    &colors->capable);
-	ret |= init_esc(colors, "sg", "30;43", &colors->setgid);
-	ret |= init_esc(colors, "su", "37;41", &colors->setuid);
+	fail = fail || init_esc(colors, "fi", "",      &colors->file);
+	fail = fail || init_esc(colors, "mh", NULL,    &colors->multi_hard);
+	fail = fail || init_esc(colors, "ex", "01;32", &colors->executable);
+	fail = fail || init_esc(colors, "ca", NULL,    &colors->capable);
+	fail = fail || init_esc(colors, "sg", "30;43", &colors->setgid);
+	fail = fail || init_esc(colors, "su", "37;41", &colors->setuid);
 
-	ret |= init_esc(colors, "di", "01;34", &colors->directory);
-	ret |= init_esc(colors, "st", "37;44", &colors->sticky);
-	ret |= init_esc(colors, "ow", "34;42", &colors->other_writable);
-	ret |= init_esc(colors, "tw", "30;42", &colors->sticky_other_writable);
+	fail = fail || init_esc(colors, "di", "01;34", &colors->directory);
+	fail = fail || init_esc(colors, "st", "37;44", &colors->sticky);
+	fail = fail || init_esc(colors, "ow", "34;42", &colors->other_writable);
+	fail = fail || init_esc(colors, "tw", "30;42", &colors->sticky_other_writable);
 
-	ret |= init_esc(colors, "ln", "01;36", &colors->link);
-	ret |= init_esc(colors, "or", NULL,    &colors->orphan);
-	ret |= init_esc(colors, "mi", NULL,    &colors->missing);
+	fail = fail || init_esc(colors, "ln", "01;36", &colors->link);
+	fail = fail || init_esc(colors, "or", NULL,    &colors->orphan);
+	fail = fail || init_esc(colors, "mi", NULL,    &colors->missing);
 	colors->link_as_target = false;
 
-	ret |= init_esc(colors, "bd", "01;33", &colors->blockdev);
-	ret |= init_esc(colors, "cd", "01;33", &colors->chardev);
-	ret |= init_esc(colors, "do", "01;35", &colors->door);
-	ret |= init_esc(colors, "pi", "33",    &colors->pipe);
-	ret |= init_esc(colors, "so", "01;35", &colors->socket);
+	fail = fail || init_esc(colors, "bd", "01;33", &colors->blockdev);
+	fail = fail || init_esc(colors, "cd", "01;33", &colors->chardev);
+	fail = fail || init_esc(colors, "do", "01;35", &colors->door);
+	fail = fail || init_esc(colors, "pi", "33",    &colors->pipe);
+	fail = fail || init_esc(colors, "so", "01;35", &colors->socket);
 
-	if (ret != 0) {
+	if (fail) {
 		goto fail;
 	}
 
@@ -764,6 +764,11 @@ static bool is_link_broken(const struct BFTW *ftwbuf) {
 	}
 }
 
+bool colors_need_stat(const struct colors *colors) {
+	return colors->setuid || colors->setgid || colors->executable || colors->multi_hard
+		|| colors->sticky_other_writable || colors->other_writable || colors->sticky;
+}
+
 /** Get the color for a file. */
 static const struct esc_seq *file_color(const struct colors *colors, const char *filename, const struct BFTW *ftwbuf, enum bfs_stat_flags flags) {
 	enum bfs_type type = bftw_type(ftwbuf, flags);
@@ -1105,11 +1110,15 @@ static int print_link_target(CFILE *cfile, const struct BFTW *ftwbuf) {
 }
 
 /** Format some colored output to the buffer. */
-BFS_FORMATTER(2, 3)
+attr(printf(2, 3))
 static int cbuff(CFILE *cfile, const char *format, ...);
 
 /** Dump a parsed expression tree, for debugging. */
-static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
+static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose, int depth) {
+	if (depth >= 2) {
+		return dstrcat(&cfile->buffer, "(...)");
+	}
+
 	if (!expr) {
 		return dstrcat(&cfile->buffer, "(null)");
 	}
@@ -1118,13 +1127,7 @@ static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
 		return -1;
 	}
 
-	const struct bfs_expr *lhs = NULL;
-	const struct bfs_expr *rhs = NULL;
-
 	if (bfs_expr_is_parent(expr)) {
-		lhs = expr->lhs;
-		rhs = expr->rhs;
-
 		if (cbuff(cfile, "${red}%pq${rs}", expr->argv[0]) < 0) {
 			return -1;
 		}
@@ -1152,21 +1155,20 @@ static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
 		}
 	}
 
-	if (lhs) {
+	int count = 0;
+	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
 		if (dstrcat(&cfile->buffer, " ") != 0) {
 			return -1;
 		}
-		if (print_expr(cfile, lhs, verbose) != 0) {
-			return -1;
-		}
-	}
-
-	if (rhs) {
-		if (dstrcat(&cfile->buffer, " ") != 0) {
-			return -1;
-		}
-		if (print_expr(cfile, rhs, verbose) != 0) {
-			return -1;
+		if (++count >= 3) {
+			if (dstrcat(&cfile->buffer, "...") != 0) {
+				return -1;
+			}
+			break;
+		} else {
+			if (print_expr(cfile, child, verbose, depth + 1) != 0) {
+				return -1;
+			}
 		}
 	}
 
@@ -1177,7 +1179,7 @@ static int print_expr(CFILE *cfile, const struct bfs_expr *expr, bool verbose) {
 	return 0;
 }
 
-BFS_FORMATTER(2, 0)
+attr(printf(2, 0))
 static int cvbuff(CFILE *cfile, const char *format, va_list args) {
 	const struct colors *colors = cfile->colors;
 	int error = errno;
@@ -1276,12 +1278,12 @@ static int cvbuff(CFILE *cfile, const char *format, va_list args) {
 					break;
 
 				case 'e':
-					if (print_expr(cfile, va_arg(args, const struct bfs_expr *), false) != 0) {
+					if (print_expr(cfile, va_arg(args, const struct bfs_expr *), false, 0) != 0) {
 						return -1;
 					}
 					break;
 				case 'E':
-					if (print_expr(cfile, va_arg(args, const struct bfs_expr *), true) != 0) {
+					if (print_expr(cfile, va_arg(args, const struct bfs_expr *), true, 0) != 0) {
 						return -1;
 					}
 					break;
