@@ -27,13 +27,22 @@ typedef __attribute__((aligned(alignof(size_t)))) char dchar;
 typedef char dchar;
 #endif
 
+/**
+ * Free a dynamic string.
+ *
+ * @param dstr
+ *         The string to free.
+ */
+void dstrfree(dchar *dstr);
+
 /**
  * Allocate a dynamic string.
  *
- * @param capacity
+ * @param cap
  *         The initial capacity of the string.
  */
-dchar *dstralloc(size_t capacity);
+attr(malloc(dstrfree, 1))
+dchar *dstralloc(size_t cap);
 
 /**
  * Create a dynamic copy of a string.
@@ -41,6 +50,7 @@ dchar *dstralloc(size_t capacity);
  * @param str
  *         The NUL-terminated string to copy.
  */
+attr(malloc(dstrfree, 1))
 dchar *dstrdup(const char *str);
 
 /**
@@ -51,6 +61,7 @@ dchar *dstrdup(const char *str);
  * @param n
  *         The maximum number of characters to copy from str.
  */
+attr(malloc(dstrfree, 1))
 dchar *dstrndup(const char *str, size_t n);
 
 /**
@@ -59,6 +70,7 @@ dchar *dstrndup(const char *str, size_t n);
  * @param dstr
  *         The dynamic string to copy.
  */
+attr(malloc(dstrfree, 1))
 dchar *dstrddup(const dchar *dstr);
 
 /**
@@ -69,6 +81,7 @@ dchar *dstrddup(const dchar *dstr);
  * @param len
  *         The length of the string, which may include internal NUL bytes.
  */
+attr(malloc(dstrfree, 1))
 dchar *dstrxdup(const char *str, size_t len);
 
 /**
@@ -86,24 +99,24 @@ size_t dstrlen(const dchar *dstr);
  *
  * @param dstr
  *         The dynamic string to preallocate.
- * @param capacity
+ * @param cap
  *         The new capacity for the string.
  * @return
  *         0 on success, -1 on failure.
  */
-int dstreserve(dchar **dstr, size_t capacity);
+int dstreserve(dchar **dstr, size_t cap);
 
 /**
  * Resize a dynamic string.
  *
  * @param dstr
  *         The dynamic string to resize.
- * @param length
+ * @param len
  *         The new length for the dynamic string.
  * @return
  *         0 on success, -1 on failure.
  */
-int dstresize(dchar **dstr, size_t length);
+int dstresize(dchar **dstr, size_t len);
 
 /**
  * Append to a dynamic string.
@@ -230,7 +243,7 @@ int dstrxcpy(dchar **dest, const char *str, size_t len);
  * @return
  *         The created string, or NULL on failure.
  */
-BFS_FORMATTER(1, 2)
+attr(printf(1, 2))
 char *dstrprintf(const char *format, ...);
 
 /**
@@ -243,7 +256,7 @@ char *dstrprintf(const char *format, ...);
  * @return
  *         The created string, or NULL on failure.
  */
-BFS_FORMATTER(1, 0)
+attr(printf(1, 0))
 char *dstrvprintf(const char *format, va_list args);
 
 /**
@@ -258,7 +271,7 @@ char *dstrvprintf(const char *format, va_list args);
  * @return
  *         0 on success, -1 on failure.
  */
-BFS_FORMATTER(2, 3)
+attr(printf(2, 3))
 int dstrcatf(dchar **str, const char *format, ...);
 
 /**
@@ -273,7 +286,7 @@ int dstrcatf(dchar **str, const char *format, ...);
  * @return
  *         0 on success, -1 on failure.
  */
-BFS_FORMATTER(2, 0)
+attr(printf(2, 0))
 int dstrvcatf(dchar **str, const char *format, va_list args);
 
 /**
@@ -306,12 +319,4 @@ int dstrescat(dchar **dest, const char *str, enum wesc_flags flags);
  */
 int dstrnescat(dchar **dest, const char *str, size_t n, enum wesc_flags flags);
 
-/**
- * Free a dynamic string.
- *
- * @param dstr
- *         The string to free.
- */
-void dstrfree(dchar *dstr);
-
 #endif // BFS_DSTRING_H
