@@ -1085,7 +1085,7 @@ bool eval_or(const struct bfs_expr *expr, struct bfs_eval *state) {
  * Evaluate the comma operator.
  */
 bool eval_comma(const struct bfs_expr *expr, struct bfs_eval *state) {
-	bool ret = uninit(ret, false);
+	bool ret uninit(false);
 
 	for (struct bfs_expr *child = bfs_expr_children(expr); child; child = child->next) {
 		ret = eval_expr(child, state);
@@ -1120,12 +1120,12 @@ static void eval_status(struct bfs_eval *state, struct bfs_bar *bar, struct time
 	const struct BFTW *ftwbuf = state->ftwbuf;
 
 	dchar *status = NULL;
-	dchar *rhs = dstrprintf(" (visited: %zu, depth: %2zu)", count, ftwbuf->depth);
+	dchar *rhs = dstrprintf(" (visited: %'zu, depth: %2zu)", count, ftwbuf->depth);
 	if (!rhs) {
 		return;
 	}
 
-	size_t rhslen = dstrlen(rhs);
+	size_t rhslen = xstrwidth(rhs);
 	if (3 + rhslen > width) {
 		dstresize(&rhs, 0);
 		rhslen = 0;
