@@ -197,14 +197,51 @@
 #define SLIST_CHECK_(list) \
 	(void)sizeof(list->tail - &list->head)
 
+/**
+ * Get the head of a singly-linked list.
+ *
+ * @param list
+ *         The list in question.
+ * @return
+ *         The first item in the list.
+ */
+#define SLIST_HEAD(list) \
+	SLIST_HEAD_((list))
+
+#define SLIST_HEAD_(list) \
+	(SLIST_CHECK_(list), list->head)
+
 /**
  * Check if a singly-linked list is empty.
  */
 #define SLIST_EMPTY(list) \
-	SLIST_EMPTY_((list))
+	(!SLIST_HEAD(list))
 
-#define SLIST_EMPTY_(list) \
-	(SLIST_CHECK_(list), !list->head)
+/**
+ * Like container_of(), but using the head pointer instead of offsetof() since
+ * we don't have the type around.
+ */
+#define SLIST_CONTAINER_(tail, head, next) \
+	(void *)((char *)tail - ((char *)&head->next - (char *)head))
+
+/**
+ * Get the tail of a singly-linked list.
+ *
+ * @param list
+ *         The list in question.
+ * @param node (optional)
+ *         If specified, use item->node.next rather than item->next.
+ * @return
+ *         The last item in the list.
+ */
+#define SLIST_TAIL(...) \
+	SLIST_TAIL_(__VA_ARGS__, )
+
+#define SLIST_TAIL_(list, ...) \
+	SLIST_TAIL__((list), LIST_NEXT_(__VA_ARGS__))
+
+#define SLIST_TAIL__(list, next) \
+	(list->head ? SLIST_CONTAINER_(list->tail, list->head, next) : NULL)
 
 /**
  * Check if an item is attached to a singly-linked list.
@@ -238,6 +275,8 @@
  *         The item to insert.
  * @param node (optional)
  *         If specified, use item->node.next rather than item->next.
+ * @return
+ *         A cursor for the next item.
  */
 #define SLIST_INSERT(list, cursor, ...) \
 	SLIST_INSERT_(list, cursor, __VA_ARGS__, )
@@ -245,11 +284,12 @@
 #define SLIST_INSERT_(list, cursor, item, ...) \
 	SLIST_INSERT__((list), (cursor), (item), LIST_NEXT_(__VA_ARGS__))
 
-#define SLIST_INSERT__(list, cursor, item, next) LIST_VOID_( \
-	bfs_assert(!SLIST_ATTACHED__(list, item, next)), \
-	item->next = *cursor, \
-	*cursor = item, \
-	list->tail = item->next ? list->tail : &item->next)
+#define SLIST_INSERT__(list, cursor, item, next) \
+	(bfs_assert(!SLIST_ATTACHED__(list, item, next)), \
+	 item->next = *cursor, \
+	 *cursor = item, \
+	 list->tail = item->next ? list->tail : &item->next, \
+	 &item->next)
 
 /**
  * Add an item to the tail of a singly-linked list.
@@ -265,7 +305,7 @@
 	SLIST_APPEND_(list, __VA_ARGS__, )
 
 #define SLIST_APPEND_(list, item, ...) \
-	SLIST_INSERT_(list, (list)->tail, item, __VA_ARGS__)
+	LIST_VOID_(SLIST_INSERT_(list, (list)->tail, item, __VA_ARGS__))
 
 /**
  * Add an item to the head of a singly-linked list.
@@ -281,7 +321,7 @@
 	SLIST_PREPEND_(list, __VA_ARGS__, )
 
 #define SLIST_PREPEND_(list, item, ...) \
-	SLIST_INSERT_(list, &(list)->head, item, __VA_ARGS__)
+	LIST_VOID_(SLIST_INSERT_(list, &(list)->head, item, __VA_ARGS__))
 
 /**
  * Add an entire singly-linked list to the tail of another.
@@ -317,15 +357,15 @@
 
 #define SLIST_REMOVE__(list, cursor, next) \
 	(list->tail = (*cursor)->next ? list->tail : cursor, \
-	 slist_remove_impl(*cursor, cursor, &(*cursor)->next, list->tail, sizeof(*cursor)))
+	 slist_remove_impl(*cursor, cursor, &(*cursor)->next, sizeof(*cursor)))
 
 // Helper for SLIST_REMOVE()
-static inline void *slist_remove_impl(void *ret, void *cursor, void *next, void *tail, size_t size) {
+static inline void *slist_remove_impl(void *ret, void *cursor, void *next, size_t size) {
 	// ret = *cursor;
 	// *cursor = ret->next;
 	memcpy(cursor, next, size);
-	// ret->next = *list->tail; (NULL)
-	memcpy(next, tail, size);
+	// ret->next = NULL;
+	memset(next, 0, size);
 	return ret;
 }
 
