@@ -1,3 +1,3 @@
-skip_unless invoke_bfs -regextype emacs -quit
+invoke_bfs -regextype emacs -quit || skip
 
 bfs_diff basic -regextype emacs -regex '.*/\(f+o?o?\|bar\)'
